export default function whichModule(exported: object) {
  if (typeof require === 'undefined') return null;
  for (
    let i = 0, files = Object.keys(require.cache), mod;
    i < files.length;
    i++
  ) {
    mod = require.cache[files[i]];
    if (mod!.exports === exported) return mod;
  }
  return null;
}
